/*
 * main.c
 *
 *  Created on: 01/10/2010
 *      Author: Saúl Zapotecas
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "global.h"
#include "io_data.h"
#include "allocate.h"
#include "rand.h"
#include "problemdef.h"
#include "scalarization.h"

#include "initialize.h"
#include "evaluate.h"
#include "dMOPSO.h"

#include "update.h"

#include "report.h"
#include "moea_tools.h"

moea_parameters moea;
mop_parameters mop;

void (*test_problem)(double*, double*, double*);
double (*objective)(double*, double*, double*);

void report_run_header(FILE *file)
{
	fprintf(file, "# ########################################\n");
	fprintf(file, "# dMOPSO\n");
	fprintf(file, "# --------------------\n");
	fprintf(file, "# population size: %d\n", moea.pop_size);
	fprintf(file, "# generations: %d\n", moea.max_gen);
	fprintf(file, "# age: %d\n", moea.age);
	fprintf(file, "# w: random(0.1, 0.5)\n");
	fprintf(file, "# C1: random(1.2, 2.0)\n");
	fprintf(file, "# C2: random(1.2, 2.0)\n");
	fprintf(file, "# archive size: no-archive\n");
	fprintf(file, "# theta (NBI): 5.0\n");
	fprintf(file, "# --------------------\n");
	fprintf(file, "# problem: %s\n", mop.str_mop_name);
	fprintf(file, "# variables: %d\n", mop.nreal);
	fprintf(file, "# objectives: %d\n", mop.nobj);
	fprintf(file, "# ########################################\n");
	return;
}

int factorial(int n)
{
	return (n <= 1) ? 1 : n * factorial(n - 1);
}
int combination(int m, int H)
{
	int N, i;
	int num;

	num = 1;
	for (i = 1; i <= m - 1; ++i)
	{
		num *= (H + i);
	}

	N = num / factorial(m - 1);
	printf("\nnum: %d\t fac(m-1): %d\n", num, factorial(m - 1));
	return N;
}

void report_population(FILE *file, population *pop, int gen)
{
	int i, j;
	fprintf(file, "# gen = %d\n", gen);
	for (i = 0; i < pop->size; i++)
	{
		for (j = 0; j < mop.nobj; j++)
		{
			fprintf(file, "%.6f ", pop->ind[i].obj[j]);
		}
		fprintf(file, "\n");
	}
	fprintf(file, "# end gen = %d\n", gen);
	fflush(file);
	return;
}

int main(int argc, char **argv)
{
	population *parent_pop;
	population *mixed_pop;
	population *gbest_pop;

	weight *W;

	FILE *gp = NULL;
	FILE *all_run = NULL;

	int gen, i;

	if (argc < 2)
	{
		printf("\n Usage ./nsga2r random_seed \n");
		exit(1);
	}

	/* Initializing seed for random numbers */
	moea.seed = (double) atof(argv[1]);
	randomize();

	/* Setting problem data */
	reading_input_data();
	/* Establishing the H value, for weighted vectors and population size */
	/* **********************************************************
	 * dMOPSO is designed only for 2 and 3 objectives. (as in the paper)
	 * 2-objectives: H=99 => population size=100
	 * 3-objectives: H=23 => population size=300
	 * **********************************************************/
	moea.H = (mop.nobj == 2) ? 99 : 23;
	moea.pop_size = combination(mop.nobj, moea.H);
	printf("H: %d\t\t N: %d", moea.H, moea.pop_size);

	printing_input_data();
	strcpy(moea.str_moea_name, "dMOPSO");

	setting_mop(mop.str_mop_name);

	/* Defining scalar problem (NBI, TCH, WEI) */
	objective = &TCH;

	/* Defining populations */
	parent_pop = (population*) malloc(sizeof(population));
	mixed_pop = (population*) malloc(sizeof(population));
	gbest_pop = (population*) malloc(sizeof(population));

	/* Defining weights */
	W = (weight*) malloc(sizeof(weight));

	/* Allocating memory for populations */
	allocate_memory_pop(parent_pop, moea.pop_size);
	allocate_memory_pop(mixed_pop, 2 * moea.pop_size);
	allocate_memory_pop(gbest_pop, moea.pop_size);

	/* Allocating memory for weights */
	allocate_memory_weight(W, moea.pop_size);

	/* Opening files */
	gp = (FILE *) popen(GNUPLOT_COMMAND, "w");
	all_run = (FILE *) fopen("all_run.dat", "w");

	/* ************************************************************************* */
	/* The Multi-Objective Evolutionary Algorithm ****************************** */
	/* ************************************************************************* */
	printf("\nMOEA has started\n");
	moea.z = malloc(sizeof(double) * mop.nobj);
	moea.n = malloc(sizeof(double) * mop.nobj);
	for (i = 0; i < mop.nobj; i++)
	{
		moea.z[i] = INF;
		moea.n[i] = _INF;
	}
	initialize_uniform_weights(W, moea.H);
	initialize_pop_LH(parent_pop);
	evaluate_pop(parent_pop);

	/*update_bently(parent_pop, nodomitated_list);*/
	get_gbest(W, parent_pop, gbest_pop);

	moea.age = 2;
	report_run_header(all_run);
	report_population(all_run, gbest_pop, 1);
	for (gen = 2; gen <= moea.max_gen; gen++)
	{
		printf("Generation: %d\n", gen);
		next_pop(W, gbest_pop, parent_pop);
		merge_population(gbest_pop, parent_pop, mixed_pop);
		get_gbest(W, mixed_pop, gbest_pop);
		report_population(all_run, gbest_pop, gen);
	}
	printf("gbest: %d\n", gbest_pop->size);

	display_pop(gbest_pop, gp, gen - 1);

	printf("\nMOEA has finished\n");
	free(moea.z);
	free(moea.n);
	/* ************************************************************************* */
	/* ************************************************************************* */

	/* Closing files */
	fclose(gp);
	fclose(all_run);

	/* Deallocating memory for populations */
	deallocate_memory_pop(parent_pop);
	deallocate_memory_pop(mixed_pop);
	deallocate_memory_pop(gbest_pop);

	free(parent_pop);
	free(mixed_pop);
	free(gbest_pop);

	/* Deallocating memory bounds */
	free(mop.low_bound);
	free(mop.up_bound);

	/* Deallocating memory for weights */
	deallocate_memory_weight(W);
	free(W);
	return EXIT_SUCCESS;
}
